#include <stdio.h>

#define PROGRAM_NAME "Zelda 64 Sequence Ripper"
#define VERSION "v0.1c"

/* Function prototypes */

int check_file (FILE* input, int pointers_offset);
unsigned int get16 (FILE *input);
unsigned int get32 (FILE *input);
int rip_sequences (FILE *input, int seqcount, int sequence_bank_offset, int pointers_offset);

/* Global variables and structs */

int seq_count;

struct sequence
{
    unsigned int start_offset;
    int lenght;
    int seq_type;   /* 0x0200, 0x0201 or 0x0202 */
} sequence_list[0x100];

/* Functions */

int check_file (FILE* input, int pointers_offset)
{
    unsigned char ch1, ch2, ch3, ch4;
    
    fseek(input, pointers_offset, SEEK_SET);
    ch1 = getc(input);
    ch2 = getc(input);
    ch3 = getc(input);
    ch4 = getc(input);

    if ( ch1 == 0 && (ch2 > 0x30) && ch3 == 0 && ch4 == 0) 
    {
        return 1;    /* valid file && pointers offset */
    }
    else 
    {
        return 0; /* wrong file or seq bank offset */
    }
}

unsigned int get32 (FILE *input)
{
    unsigned int byte1, byte2, byte3, byte4;
    unsigned int unsigned32;

    byte1 = fgetc(input);
    byte2 = fgetc(input);
    byte3 = fgetc(input);
    byte4 = fgetc(input);  
    
    unsigned32 = (byte1 << 24) + (byte2 << 16) + (byte3 << 8) + byte4;              
    return unsigned32;
}

unsigned int get16 (FILE *input)
{
    unsigned int byte1, byte2;
    unsigned int unsigned16;

    byte1 = fgetc(input);
    byte2 = fgetc(input);
    
    unsigned16 = (byte1 << 8) + byte2;
    return unsigned16;
}

int rip_sequences (FILE *input, int seqcount, int sequence_bank_offset, int pointers_offset)
{
    char filename[14];
    char ch;
    int i, j;
    unsigned int offset;
  
   fseek(input, ((pointers_offset) + 0x10), SEEK_SET);  
  
    for (i = 0; i < seqcount; i++)
    {
        fseek(input, ((pointers_offset) + 0x10 + (0x10 * i) ), SEEK_SET); 
        sequence_list[i].start_offset = get32(input);
        sequence_list[i].lenght = get32(input);
        sequence_list[i].seq_type = get16(input);
        get16(input);   /* padding */
        get32(input);   /* padding */
    }

    for (i = 0; i < seqcount; i++)
    {
        sprintf(filename, "seq_%02x_%04x.zseq", i, sequence_list[i].seq_type);
        
        FILE *ripped_file = fopen( filename , "w+b"); 
        fseek(input, (sequence_bank_offset + sequence_list[i].start_offset) , SEEK_SET);        
   
        for ( j = 0; j < sequence_list[i].lenght; j++)
        {
            ch = getc(input);
            putc(ch, ripped_file);
        }
        
       fclose(ripped_file);
    }

}


/* entry-point */

int main ( int argc, char *argv[] )
{
    
    unsigned int sbank_offset = 0;
    unsigned int seq_pointers_offset = 0;        
    
    if ( argc != 4 )
    {
        printf("\n%s %s by messiaen", PROGRAM_NAME, VERSION);
        printf("\nUsage: z64seqrip.exe [ROM] [seq_bank_offset] [seq_pointer_table_offset]" );
        printf("\nExample: z64seqrip.exe zelootma.z64 0x44DF0 0xBCC6A0 (Debug ROM)" );
        printf("\n         z64seqrip.exe mmask.z64 0x46AF0 0xBC7690 (Majora's Mask uncompressed)\n");
    } 
    else 
    {
        sbank_offset = strtoul (argv[2],NULL,0);
        seq_pointers_offset = strtoul (argv[3],NULL,0);

        FILE *input_file = fopen( argv[1], "rb" );
   
        if (!(input_file == 0 ))
        {
            if (!( check_file(input_file, seq_pointers_offset) == 0) )
            {                 
                fseek(input_file, (seq_pointers_offset + 1), SEEK_SET);
                seq_count = fgetc(input_file);   
                printf("\n%s %s by messiaen", PROGRAM_NAME, VERSION);         
                printf("\nRipping %d sequences ... ", seq_count);
                rip_sequences(input_file, seq_count, sbank_offset, seq_pointers_offset);
                printf("done! \n");
            }
            else
            {
                printf ("\nNot a valid pointer table offset.\n");
            } 
        }
        else
        {
            printf ("\nError opening input file.\n");
        }      
    }   
}
